//	Author:		Dario Accornero
//	BeDev ID:	10183
//	Mailto:		adario@cs.bu.edu
//	Project:	BeScript
//	Module:
//	Scripting targets implementation.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

//	System headers.

#ifndef _APPLICATION_H
#include <Application.h>
#endif
#ifndef _STDIO_H_
#include <stdio.h>
#endif
#ifndef	_STRING_H_
#include <string.h>
#endif
#ifndef BESCRIPT
#ifndef _OUTLINE_LIST_VIEW_H
#include <OutlineListView.h>
#endif
#endif	// !BESCRIPT

//	Local headers.

#include "ScriptingElement.h"
#include "ScriptingTargets.h"
#ifndef BESCRIPT
#include "ScriptingUtils.h"
#include "BeScriptPeekApp.h"
#endif

//	Utilities and local declarations.

namespace {

	//	Paranoia mode on: verify that a string isn't made of whitespace only.
	
	bool ContainsValidChars(const BString& str);
	bool ContainsValidChars(const char* str);
	
	//	Icons for targets.
	
	#ifndef BESCRIPT
	BBitmap* stApplicationIcon = 0;
	BBitmap* stWindowIcon = 0;
	BBitmap* stViewIcon = 0;
	#endif
};

//	FUNCTIONS:
//	Build list of running applications.
//	Items of the list are of type "ScriptingApplication*", but can be
//	safely retrieved as "ScriptingElement*" since the former inherits from the latter.

BList*	GetRunningApplications(const char* currentSignature)
{
	//	Traverse list of running teams returned by the system roster.
	//	This is derived from Attila Mezei's "hey" code.

	BList team_list;
	be_roster->GetAppList(&team_list);
	int32 items = team_list.CountItems();
	BList* app_list = new BList(items - 1);
	for (int32 i = 0; i < items; ++i)
	{
		//	Retrieve team ID and application runtime info.
		
		team_id	teamid;
		teamid = (team_id)(team_list.ItemAt(i));
		app_info appinfo;
		be_roster->GetRunningAppInfo(teamid, &appinfo);

		//	We don't operate on ourselves -- we're not masochistic.

		if (!strcmp(appinfo.signature, currentSignature))
			continue;

		//	Add application entry to our list.
		//	Also, ask application for its supported suites.

		ScriptingApplication* app = new ScriptingApplication(	BString(appinfo.ref.name),
																appinfo,
																new BMessenger(appinfo.signature));
		app_list->AddItem(app);
	}
	return	app_list;
}

//	Creates an instance of a scriptable application.
//	It locates the application (if available) in the system roster, and returns
//	a complete scriptable object.

ScriptingApplication*	CreateScriptingApplication(const char* theAppIdentifier)
{
	//	Traverse list of running teams returned by the system roster.
	//	This is derived from Attila Mezei's "hey" code.

	bool found = false;
	BList team_list;
	be_roster->GetAppList(&team_list);
	int32 items = team_list.CountItems();
	ScriptingApplication* app = 0;
	for (int32 i = 0; !found && i < items; ++i)
	{
		//	Retrieve team ID and application runtime info.
		
		team_id	teamid;
		teamid = (team_id)team_list.ItemAt(i);
		app_info appinfo;
		be_roster->GetRunningAppInfo(teamid, &appinfo);

		//	Add application entry to our list.
		//	Also, ask application for its supported suites. This is done
		//	automatically by the ScriptingApplication constructor.

		if (!strcmp(theAppIdentifier, appinfo.ref.name) ||
			!strcmp(theAppIdentifier, appinfo.signature))
		{
			app = new ScriptingApplication(	BString(appinfo.ref.name),
											appinfo,
											new BMessenger(appinfo.signature));
			found = true;
		}
	}	
	return	app;
}

//	CLASSES:
//	--- ScriptingApplication:
//		Constructor.

ScriptingApplication::ScriptingApplication(	const BString& theName,
											const app_info& theAppInfo,
											BMessenger* theMessenger,
											ScriptingElement* theParent)
: ScriptingElement(theMessenger, theParent)
{
	//	Initialize object and check whether it is actually scriptable.
	//	This is done by simply asking the object for its supported suites.
	
	fAppInfo = theAppInfo;
	fInternalName.SetTo(fAppInfo.signature);
	fName = theName;
	fWindows = 0;
	fNumber = 0;
	status_t result = FindSuites();
	if (result == B_OK)
	{
		//	Verify scriptability and assign string name to item.
		
		fIsScriptable = fSuites.HasString(SF_SUITES);
		#ifndef BESCRIPT
		UpdateItem();
		#endif
	}
}

//	--- ScriptingApplication:
//		Destructor.

ScriptingApplication::~ScriptingApplication()
{
	//	Destroy messenger. We know we can do this because we retain ownership of the object.
	
	delete fMessenger;
}

//	--- ScriptingApplication:
//		Display element on the console.

void	ScriptingApplication::PrintToStream() const
{
	printf("\"%s\" (%s)", fInternalName.String(), fName.String());
	if (fIsScriptable)
	{
		//	Display all available suites.
		
		printf(" supports:\n");
		int32 suitesFound = 0;
		const char* suiteName;
		while (fSuites.FindString(SF_SUITES, suitesFound++, &suiteName) == B_OK)
		{
			if (suitesFound > 1)
				printf(" ");
			printf("%s", suiteName);
		}
		printf("\n");
	
		//	Display all available windows.
		//	This is done by calling the default display method for an element.
		
		ScriptingElement::PrintToStream();
	}
	else
		printf(" is not scriptable\n");
}

//	--- ScriptingApplication:
//		Analyze an application.

int32	ScriptingApplication::Analyze(int32 theLevel)
{
	//	Safety check: nothing to analyze for a non-scriptable application.
	
	if (!IsScriptable())
		return 0;
	
	//	Count number of application windows.
	//	Note that applications do not have a base specifier access message.
	
	BMessage countWindowMsg(B_COUNT_PROPERTIES);
	countWindowMsg.AddSpecifier(SM_WINDOW);
	BMessage reply;
	status_t result = fMessenger->SendMessage(&countWindowMsg, &reply);
	if (result != B_OK)
	{
		#ifndef BESCRIPT
		fprintf(stderr, "cannot 'count Window' of application '%s' (%s)\n", fName.String(), strerror(result));
		#endif
		return 0;
	}
	result = reply.FindInt32(SF_RESULT, &fWindows);
	if (result != B_OK)
	{
		#ifndef BESCRIPT
		fprintf(stderr, "cannot extract number of windows for application '%s' (%s)\n", fName.String(), strerror(result));
		#endif
		return 0;
	}
	
	//	Prepare traversal. All windows have a level ID of 1.
	
	reply.MakeEmpty();
	++theLevel;
	
	//	Create scripting windows list.
	
	#if DEBUG
	printf("Application %s reports %ld windows\n", fName.String(), fWindows);
	#endif
	fChildren = new BList(fWindows);
	for (int32 index = 0; index < fWindows; ++index)
	{		
		//	Create window, analyze its contents, and add it to the application.
		//	We'll create a messenger per window because windows retain messenger ownership.
		
		ScriptingWindow* window = new ScriptingWindow(index, new BMessenger(*fMessenger), this, theLevel);
		if (window->IsScriptable())
		{
			window->Analyze(theLevel);
			fChildren->AddItem(window);
		}
		else
			delete window;
	}
	fWindows = fChildren->CountItems();
	
	//	Send message to application, signaling the traversal is over.
	
	#ifndef BESCRIPT
	void* object = this;
	BMessage doneMsg(BESCRIPTPEEK_ADD_APPLICATION);
	doneMsg.AddPointer(SF_RESULT, object);
	be_app->PostMessage(&doneMsg);
	#endif
	
	return fWindows;
}

//	--- ScriptingApplication:
//		Analyzis thread creation.

int32 ScriptingApplication::AnalyzeApplication()
{
	BString threadName(fName);
	threadName += " analysis";
	thread_id analyzeThread = spawn_thread( (thread_func)AnalyzeThread,
											threadName.String(),
											B_NORMAL_PRIORITY,
											this);
	if (analyzeThread >= 0)
		return resume_thread(analyzeThread);
	else
		return Analyze();
}

//	--- ScriptingApplication:
//		Analysis workhorse thread.

int32 ScriptingApplication::AnalyzeThread(ScriptingApplication* app)
{
	return app->Analyze();
}

//	--- ScriptingApplication:
//		Retrieve dynamic name.

status_t ScriptingApplication::FindDynamicName()
{
	//	We ask for a "Name".
	
	BMessage reply(B_NO_REPLY);
	BMessage nameMsg(B_GET_PROPERTY);
	nameMsg.AddSpecifier(SM_NAME);
	AddAccess(&nameMsg);
	status_t result = fMessenger->SendMessage(&nameMsg, &reply);
	if (result == B_OK)
	{
		const char* name;
		result = reply.FindString(SF_RESULT, &name);
		if (result == B_OK)
			if (fName != name)
				fName.SetTo(name);
	}
	return result;
}

//	--- ScriptingApplication:
//		Update item description.
#ifndef BESCRIPT
void ScriptingApplication::UpdateItem()
{
	BString text("\"");
	text += fInternalName;
	text += "\"";
	SetText(text.String());
	if (!fIsScriptable)
		SetDynamicText("NOT SCRIPTABLE");
	else
		SetDynamicText(fName.String());
}

//	--- ScriptingApplication:
//		Return bitmap.
const BBitmap* ScriptingApplication::GetBitmap() const
{
	if (!stApplicationIcon)
	{
		app_info aInfo;
		status_t error = be_app->GetAppInfo(&aInfo);
		if (error != B_OK)
		{
			BRect rect(0, 0, 15, 15);
			stApplicationIcon = new BBitmap(rect, B_CMAP8);
		}
		else
		{
			BFile file(&aInfo.ref, B_READ_ONLY);
			stApplicationIcon = LoadBitmapFromResource(file, 19);
		}
	}
	return stApplicationIcon;
}
#endif	// !BESCRIPT

//	--- ScriptingApplication:
//		Rescan to see if windows have been shown or hidden.

void ScriptingApplication::Rescan(BView* view)
{
	//	First of all, let's see how many windows we have.
	
	BMessage reply;
	BMessage countWindowMsg(B_COUNT_PROPERTIES);
	countWindowMsg.AddSpecifier(SM_WINDOW);
	status_t result = fMessenger->SendMessage(&countWindowMsg, &reply);
	if (result != B_OK)
		return;
	int32 windows;
	result = reply.FindInt32(SF_RESULT, &windows);
	if (result != B_OK)
		return;
	
	//	Iterate over available windows.
	
	fWindows = fChildren->CountItems();
	int32 last = MIN(fWindows, windows);
	for (int32 index = 0; index < last; ++index)
	{
		//	Obtain current window title.
		
		BMessage titleMsg(B_GET_PROPERTY);
		titleMsg.AddSpecifier(SM_TITLE);
		titleMsg.AddSpecifier(SM_WINDOW, index);
		result = fMessenger->SendMessage(&titleMsg, &reply);
		if (result != B_OK)
			break;
		const char* title;
		result = reply.FindString(SF_RESULT, &title);
		if (result != B_OK)
			break;
		
		//	Make sure this window has changed before proceeding: we sense a window change
		//	when the title has been modified. This should be enough but isn't totally accurate.
		
		ScriptingWindow* window = (ScriptingWindow*)fChildren->ItemAt(index);
		if (!window)
			break;
		const char* windowName = window->GetName();
		if (!windowName || !strlen(windowName))
			windowName = window->GetInternalName().String();
		if (!windowName)
			continue;
		if (strcmp(windowName, title) == 0)
			continue;
		
		//	We might have to modify the list displayed on screen as well.

		int32 itemIndex = -1;
		if (view)
		{
			#ifndef BESCRIPT
			BOutlineListView* listView = dynamic_cast<BOutlineListView*>(view);
			if (listView)
				itemIndex = listView->FullListIndexOf(window);
			#endif
		}
		
		//	Create new window and add it to the hierarchy.

		delete window;
		int32 theLevel = fLevel + 1;
		window = new ScriptingWindow(index, new BMessenger(*fMessenger), this, theLevel);
		if (window->IsScriptable())
		{
			window->Analyze(theLevel);
			fChildren->ReplaceItem(index, window);
			if (view && itemIndex != -1)
			{
				#ifndef BESCRIPT
				BOutlineListView* listView = (BOutlineListView*)view;
				listView->ReplaceItem(itemIndex, window);
				window->AddUnder(listView, this);
				#endif
			}
		}
		else
		{
			//	Non-scriptable windows are removed.
			
			delete window;
			fChildren->RemoveItem(index);
			--index;
			--last;
		}
	}
	
	//	Now we have to see if we need to purge or add windows.

	if (fWindows > windows)
	{
		//	We have to remove all windows that have ceased to exist in the remote application.
		//	We might have to do it also for the onscreen list.
		
		for (int32 i = windows; i < fWindows; ++i)
		{
			ScriptingWindow* window = (ScriptingWindow*)fChildren->ItemAt(i);
			if (!window)
				break;
			if (view)
			{
				#ifndef BESCRIPT
				BOutlineListView* listView = (BOutlineListView*)view;
				listView->RemoveItem(window);
				#endif
			}
			delete window;
		}
		fChildren->RemoveItems(windows, fWindows - windows);
	}
	else if (fWindows < windows)
	{
		//	The onscreen list must be handled with care: we have to add the new windows
		//	under the current application.
		
		#ifndef BESCRIPT
		BOutlineListView* listView = 0;
		if (view)
			listView = (BOutlineListView*)view;
		#endif
		for (int32 i = fWindows; i < windows; ++i)
		{
			int32 theLevel = fLevel + 1;
			ScriptingWindow* window = new ScriptingWindow(i, new BMessenger(*fMessenger), this, theLevel);
			if (window->IsScriptable())
			{
				window->Analyze(theLevel);
				fChildren->AddItem(window);
				#ifndef BESCRIPT
				if (listView)
					window->AddUnder(listView, this);
				#endif
			}
			else
				delete window;
		}
	}
	
	//	Update the object with the current number of windows.
	
	fWindows = fChildren->CountItems();
}

//	--- ScriptingWindow:
//		Constructor.

ScriptingWindow::ScriptingWindow(	const int32 theNumber,
									BMessenger* theMessenger,
									ScriptingElement* theParent,
									const int32 theLevel,
									bool addSpecifier)
: ScriptingElement(theMessenger, theParent, theLevel)
{
	//	Initialize object.
	
	fNumber = theNumber;
	fName = "";
	InitObject(fNumber, addSpecifier);
}

//	--- ScriptingWindow:
//		Constructor.

ScriptingWindow::ScriptingWindow(	const char* theName,
									BMessenger* theMessenger,
									ScriptingElement* theParent,
									const int32 theLevel,
									bool addSpecifier)
: ScriptingElement(theMessenger, theParent, theLevel)
{
	//	Initialize object.
	
	fName = theName;
	fNumber = -1;
	InitObject(fName.String(), addSpecifier);
}

//	--- ScriptingWindow:
//		Destructor.

ScriptingWindow::~ScriptingWindow()
{
	//	Destroy messenger. We know we can do this because we retain ownership of the object.
	
	delete fMessenger;
}

//	--- ScriptingWindow:
//		Object initialization.

void ScriptingWindow::InitObject(const int32 theNumber, bool addSpecifier)
{
	//	Extract window name (if valid messenger is passed).
	
	if (fMessenger && fMessenger->IsValid())
	{
		//	Base specifier access message for a window.
		//	This works under the assumption that the target messenger belongs to a BApplication.
		
		fAccess.MakeEmpty();
		if (addSpecifier)
			fAccess.AddSpecifier(SM_WINDOW, theNumber);
		
		//	Perform object initialization.
		
		InitObject();
	}
}

//	--- ScriptingWindow:
//		Object initialization.

void ScriptingWindow::InitObject(const char* theName, bool addSpecifier)
{
	//	Extract window name (if valid messenger is passed).
	
	if (fMessenger && fMessenger->IsValid())
	{
		//	Base specifier access message for a window.
		//	This works under the assumption that the target messenger belongs to a BApplication.
		
		fAccess.MakeEmpty();
		if (addSpecifier)
			fAccess.AddSpecifier(SM_WINDOW, theName);
		
		//	Perform object initialization.
		
		InitObject();
	}
}

//	--- ScriptingWindow:
//		Internal object initialization.

void ScriptingWindow::InitObject()
{
	//	Determine supported suites.

	status_t result = FindSuites();
	if (result == B_OK)
		fIsScriptable = fSuites.HasString(SF_SUITES);
	if (!fIsScriptable)
		return;

	//	Try "get InternalName".
	
	BMessage reply;
	BMessage nameMsg(B_GET_PROPERTY);
	nameMsg.AddSpecifier(SM_INTERNALNAME);
	AddAccess(&nameMsg);
	result = fMessenger->SendMessage(&nameMsg, &reply);
	if (result == B_OK)
	{
		const char* windowName;
		result = reply.FindString(SF_RESULT, &windowName);
		if (result == B_OK)
			fInternalName.SetTo(windowName);
	}

	//	Extract window name if available and possibly update it.
	
	if (ScriptingElement::DynamicNames)
		FindDynamicName();

	//	Check names and assign strings to item.
	
	#ifndef BESCRIPT
	UpdateItem();
	#endif
}

//	--- ScriptingWindow:
//		Display element on the console.

void	ScriptingWindow::PrintToStream() const
{
	IndentToStream();
	if (fName.Length() > 0 && ContainsValidChars(fName))
		printf("%s [%ld]\n", fName.String(), fNumber);
	else
		printf("%s [%ld]\n", fInternalName.String(), fNumber);
	ScriptingElement::PrintToStream();
}

//	--- ScriptingWindow:
//		Analyze window and all of its views.

int32	ScriptingWindow::Analyze(int32 theLevel)
{
	//	Safety check: nothing to analyze for a non-scriptable window.
	
	if (!IsScriptable())
	{
		#ifndef BESCRIPT
		BString name;
		if (fInternalName.Length() > 0 && ContainsValidChars(fInternalName))
			SetText(fInternalName.String());
		else
		{
			if (fName.Length() > 0 && ContainsValidChars(fName))
				SetText(fName.String());
			else
			{
				char buf[64];
				sprintf(buf, "Window %ld", fNumber);
				SetText(buf);
			}
		}
		SetDynamicText("NOT SCRIPTABLE");
		#endif
		return 0;
	}
	
	//	Count number of views for this window.
	
	BMessage reply;
	BMessage countViewMsg(B_COUNT_PROPERTIES);
	countViewMsg.AddSpecifier(SM_VIEW);
	AddAccess(&countViewMsg);
	status_t result = fMessenger->SendMessage(&countViewMsg, &reply);
	if (result != B_OK)
	{
		#ifndef BESCRIPT
		if (fNumber >= 0)
			fprintf(stderr, "cannot 'count View' of Window %ld of %s (%s)\n",
					fNumber, ParentName(), strerror(result));
		else
			fprintf(stderr, "cannot 'count View' of Window %s of %s (%s)\n",
					fName.String(), ParentName(), strerror(result));
		#endif
		return 0;
	}
	int32 numViews;
	result = reply.FindInt32(SF_RESULT, &numViews);
	if (result != B_OK)
	{
		#ifndef BESCRIPT
		if (fNumber >= 0)
			fprintf(stderr, "cannot extract number of views of Window %ld of %s (%s)\n",
					fNumber, ParentName(), strerror(result));
		else
			fprintf(stderr, "cannot extract number of views of Window %s of %s (%s)\n",
					fName.String(), ParentName(), strerror(result));
		#endif
		return 0;
	}
	
	//	Prepare traversal.
	
	reply.MakeEmpty();
	++theLevel;
	
	//	Create scripting views list.
	
	#if DEBUG
	printf("  Window %s of application %s reports %ld views\n", fName.String(), ParentName(), numViews);
	#endif
	fChildren = new BList(numViews);
	for (int32 index = 0; index < numViews; ++index)
	{
		//	Create view, analyze its contents, and add it to the parent window.
		
		ScriptingView* view = new ScriptingView(index, fMessenger, this, theLevel);
		if (view->IsScriptable())
		{
			view->Analyze(theLevel);
			fChildren->AddItem(view);
		}
		else
			delete view;
	}
	numViews = fChildren->CountItems();
	return numViews;
}

//	--- ScriptingWindow:
//		Retrieve dynamic name.

status_t ScriptingWindow::FindDynamicName()
{
	//	We ask for a "Title".
	
	BMessage reply(B_NO_REPLY);
	BMessage nameMsg(B_GET_PROPERTY);
	nameMsg.AddSpecifier(SM_TITLE);
	AddAccess(&nameMsg);
	status_t result = fMessenger->SendMessage(&nameMsg, &reply);
	if (result == B_OK)
	{
		const char* name;
		result = reply.FindString(SF_RESULT, &name);
		if (result == B_OK)
			if (fName != name)
				fName.SetTo(name);
	}
	return result;
}

//	--- ScriptingWindow:
//		Update item description.
#ifndef BESCRIPT
void ScriptingWindow::UpdateItem()
{
	if (fName.Length() > 0 && ContainsValidChars(fName))
		SetDynamicText(fName.String());
	if (fInternalName.Length() > 0 && ContainsValidChars(fInternalName))
		SetText(fInternalName.String());
	else if (fNumber >= 0)
	{
		char buf[64];
		sprintf(buf, "Window %ld", fNumber);
		SetText(buf);
	}
	else
		SetText("(Ghost Window)");
}

//	--- ScriptingWindow:
//		Return bitmap.
const BBitmap* ScriptingWindow::GetBitmap() const
{
	if (!stWindowIcon)
	{
		app_info aInfo;
		status_t error = be_app->GetAppInfo(&aInfo);
		if (error != B_OK)
		{
			BRect rect(0, 0, 15, 15);
			stWindowIcon = new BBitmap(rect, B_CMAP8);
		}
		else
		{
			BFile file(&aInfo.ref, B_READ_ONLY);
			stWindowIcon = LoadBitmapFromResource(file, 20);
		}
	}
	return stWindowIcon;
}
#endif

//	--- ScriptingView:
//		Constructor.

ScriptingView::ScriptingView(	const int32 theNumber,
								BMessenger* theMessenger,
								ScriptingElement* theParent,
								const int32 theLevel)
: ScriptingElement(theMessenger, theParent, theLevel)
{
	//	Assign view number and extract name (if valid messenger is passed).
	
	fNumber = theNumber;
	fName = "";
	InitObject(fNumber);
}

//	--- ScriptingView:
//		Constructor.

ScriptingView::ScriptingView(	const char* theName,
								BMessenger* theMessenger,
								ScriptingElement* theParent,
								const int32 theLevel)
: ScriptingElement(theMessenger, theParent, theLevel)
{
	//	Assign view number and extract name (if valid messenger is passed).
	
	fName = theName;
	fNumber = -1;
	InitObject(fName.String());
}

//	--- ScriptingView:
//		Destructor.

ScriptingView::~ScriptingView()
{
	//	Does nothing.
}

//	--- ScriptingView:
//		Object initialization.

void ScriptingView::InitObject(const int32 theNumber)
{
	//	Extract view name (if valid messenger is passed).
	
	if (fMessenger && fMessenger->IsValid())
	{
		//	Set up base specifier access message.
		
		fAccess.MakeEmpty();
		fAccess.AddSpecifier(SM_VIEW, theNumber);
		if (fParent)
			fParent->AddAccess(&fAccess);
		
		InitObject();
	}
}

//	--- ScriptingView:
//		Object initialization.

void ScriptingView::InitObject(const char* theName)
{
	//	Extract view name (if valid messenger is passed).
	
	if (fMessenger && fMessenger->IsValid())
	{
		//	Set up base specifier access message.
		
		fAccess.MakeEmpty();
		fAccess.AddSpecifier(SM_VIEW, theName);
		if (fParent)
			fParent->AddAccess(&fAccess);
		
		InitObject();
	}
}

//	--- ScriptingView:
//		Object initialization.

void ScriptingView::InitObject()
{
	//	Determine supported suites.

	status_t result = FindSuites();
	if (result == B_OK)
		fIsScriptable = fSuites.HasString(SF_SUITES);
	if (!fIsScriptable)
		return;

	//	Try to "get InternalName".
	//	Extract view name if available.
	
	BMessage reply;
	BMessage nameMsg(B_GET_PROPERTY);
	nameMsg.AddSpecifier(SM_INTERNALNAME);
	AddAccess(&nameMsg);
	result = fMessenger->SendMessage(&nameMsg, &reply);
	const char* viewName;
	if (result == B_OK)
	{
		result = reply.FindString(SF_RESULT, &viewName);
		if (result == B_OK)
			fInternalName.SetTo(viewName);
	}
	
	//	Now try to get a more descriptive name, such as a Label.
	
	if (ScriptingElement::DynamicNames)
		FindDynamicName();

	//	Let's see what we have.
	
	if ((fName.Length() < 1 || !ContainsValidChars(fName))
		&& (fInternalName.Length() < 1 || !ContainsValidChars(fInternalName)))
	{
		//	We really want to avoid displaying "View #" because it's a shame.
		//	One last check -- the view might be a MenuBar.
		
		bool found = false;
		int32 index = 0;
		const char* suiteName;
		while (!found && fSuites.FindString(SF_SUITES, index++, &suiteName) == B_OK)
		{
			if (!strcmp(suiteName, "suite/vnd.Be-menu"))
			{
				fInternalName.SetTo("MenuBar");
				found = true;
			}
		}
	}
	
	//	Assign strings to item.
	
	#ifndef BESCRIPT
	UpdateItem();
	#endif
}

//	--- ScriptingView:
//		Display element on the console.

void	ScriptingView::PrintToStream() const
{
	//	This method is currently identical to ScriptingWindow::PrintToStream().
	//	In the future it'll provide more details about the view.

	IndentToStream();
	if (fName.Length() > 0 && ContainsValidChars(fName))
		printf("%s [%ld]\n", fName.String(), fNumber);
	else
		printf("%s [%ld]\n", fInternalName.String(), fNumber);
	ScriptingElement::PrintToStream();
}

//	--- ScriptingView:
//		Analyze view and all of its children.

int32	ScriptingView::Analyze(int32 theLevel)
{
	//	Safety check: nothing to analyze for a non-scriptable view.
	//	Note that this should never happen.
	
	if (!IsScriptable())
	{
		#ifndef BESCRIPT
		if (fInternalName.Length() > 0 && ContainsValidChars(fInternalName))
			SetText(fInternalName.String());
		else if (fNumber >= 0)
		{
			char buf[64];
			sprintf(buf, "View %ld", fNumber);
			SetText(buf);
		}
		else
		{
			if (!(fName.Length() > 0 && ContainsValidChars(fName)))
				SetText("(Ghost View)");
		}
		SetDynamicText("NOT SCRIPTABLE");
		#endif
		return 0;
	}
	
	//	Count number of children views for this view.
	
	BMessage reply;
	BMessage countViewMsg(B_COUNT_PROPERTIES);
	countViewMsg.AddSpecifier(SM_VIEW);
	AddAccess(&countViewMsg);
	status_t result = fMessenger->SendMessage(&countViewMsg, &reply);
	if (result != B_OK)
	{
		#ifndef BESCRIPT
		if (fNumber >= 0)
			fprintf(stderr, "cannot 'count View' of View %ld of %s (%s)\n",
					fNumber, ParentName(), strerror(result));
		else
			fprintf(stderr, "cannot 'count View' of View %s of %s (%s)\n",
					fName.String(), ParentName(), strerror(result));
		#endif
		return 0;
	}
	int32 numViews;
	result = reply.FindInt32(SF_RESULT, &numViews);
	if (result != B_OK)
	{
		#ifndef BESCRIPT
		if (fNumber >= 0)
			fprintf(stderr, "cannot extract number of views of View %ld of %s (%s)\n",
					fNumber, ParentName(), strerror(result));
		else
			fprintf(stderr, "cannot extract number of views of View %s of %s (%s)\n",
					fName.String(), ParentName(), strerror(result));
		#endif
		return 0;
	}
	
	//	Prepare traversal.
	
	reply.MakeEmpty();
	++theLevel;
	
	//	Create scripting views list.
	
	#if DEBUG
	for (int i = 0; i < theLevel; ++i)
		printf("  ");
	printf("View %ld (%s,%s) reports %ld views\n", fNumber, fInternalName.String(), fName.String(), numViews);
	#endif
	fChildren = new BList(numViews);
	for (int32 index = 0; index < numViews; ++index)
	{
		//	Create view, analyze its contents, and add it to the parent view.
		
		ScriptingView* view = new ScriptingView(index, GetMessenger(), this, theLevel);
		if (view->IsScriptable())
		{
			view->Analyze(theLevel);
			fChildren->AddItem(view);
		}
		else
			delete view;
	}
	numViews = fChildren->CountItems();
	return numViews;
}

//	--- ScriptingView:
//		Update item description.
#ifndef BESCRIPT
void ScriptingView::UpdateItem()
{
	if (fName.Length() > 0 && ContainsValidChars(fName))
		SetDynamicText(fName.String());
	if (fInternalName.Length() > 0 && ContainsValidChars(fInternalName))
		SetText(fInternalName.String());
	else if (fNumber >= 0)
	{
		char buf[64];
		sprintf(buf, "View %ld", fNumber);
		SetText(buf);
	}
	else
	{
		if (!(fName.Length() > 0 && ContainsValidChars(fName)))
			SetText("(Ghost View)");
	}
}

//	--- ScriptingView:
//		Return bitmap.
const BBitmap* ScriptingView::GetBitmap() const
{
	if (!stViewIcon)
	{
		app_info aInfo;
		status_t error = be_app->GetAppInfo(&aInfo);
		if (error != B_OK)
		{
			BRect rect(0, 0, 15, 15);
			stViewIcon = new BBitmap(rect, B_CMAP8);
		}
		else
		{
			BFile file(&aInfo.ref, B_READ_ONLY);
			stViewIcon = LoadBitmapFromResource(file, 21);
		}
	}
	return stViewIcon;
}
#endif

//	Utilities:
//	Check whether a string contains only whitespace.

namespace {

#include <ctype.h>
bool ContainsValidChars(const BString& str)
{
	return ContainsValidChars(str.String());
}
bool ContainsValidChars(const char* str)
{
	int len = strlen(str);
	for (int32 index = 0; index < len; ++index)
		if (!isspace(str[index]))
			return true;
	return false;
}

};
