#include "PrefilledBitmap.h"

//	A prefilled bitmap class, containing regular and color masks, optionally read from a resource file.
//	Note that bitmaps are assumed to be in 8-bit mode (for mask generation purposes).

PrefilledBitmap::PrefilledBitmap(	BRect bounds,
									color_space space,
									const void *data,
									bool acceptsViews,
									bool needsContiguousMemory,
									bool createMasks)
: BBitmap(bounds, space, acceptsViews, needsContiguousMemory)
{
	//	We assume we'll be using 8-bit (256 color) icons with the BeOS system palette.

	int32 length = ((int32(bounds.right-bounds.left)+3) / 4) * 4;
	length *= int32(bounds.bottom-bounds.top)+1;
	SetBits(data, length, 0, space);

	//	Generate needed masks.
	
	mask = redMask = greenMask = 0;
	if (createMasks)
		CreateMasks(length, acceptsViews, needsContiguousMemory);	
}

PrefilledBitmap::PrefilledBitmap(BMessage* archive, bool acceptsViews, bool needsContiguousMemory, bool createMasks)
: BBitmap(archive)
{
	mask = redMask = greenMask = 0;
	if (createMasks)
		CreateMasks(BitsLength(), acceptsViews, needsContiguousMemory);
}

//	Make sure we removed the masks before killing this icon.

PrefilledBitmap::~PrefilledBitmap()
{
	delete greenMask;
	delete redMask;
	delete mask;
}

//	Create masks for this bitmap.

void PrefilledBitmap::CreateMasks(int32 length, bool acceptsViews, bool needsContiguousMemory)
{
	//	Create black mask.
	
	BRect bounds(Bounds());
	color_space space = ColorSpace();
	
	mask = new BBitmap(bounds, space, acceptsViews, needsContiguousMemory);
	mask->Lock();
	uint8 *maskBits = (uint8*)mask->Bits();
	uint8 *bits = (uint8*)Bits();
	for (int32 index = 0; index < length; ++index, ++maskBits, ++bits)
	{
		if (*bits != B_TRANSPARENT_8_BIT)
			*maskBits = 1;						//	Almost black in BeOS palette.
		else
			*maskBits = B_TRANSPARENT_8_BIT;
	}
	mask->Unlock();

	//	Create red mask.

	redMask = new BBitmap(bounds, space, acceptsViews, needsContiguousMemory);
	redMask->Lock();
	maskBits = (uint8*)redMask->Bits();
	bits = (uint8*)Bits();
	for (int32 index = 0; index < length; ++index, ++maskBits, ++bits)
	{
		if (*bits != B_TRANSPARENT_8_BIT)
			*maskBits = 0x2A;					//	Pure red in BeOS palette.
		else
			*maskBits = B_TRANSPARENT_8_BIT;
	}
	redMask->Unlock();

	//	Create green mask.

	greenMask = new BBitmap(bounds, space, acceptsViews, needsContiguousMemory);
	greenMask->Lock();
	maskBits = (uint8*)greenMask->Bits();
	bits = (uint8*)Bits();
	for (int32 index = 0; index < length; ++index, ++maskBits, ++bits)
	{
		if (*bits != B_TRANSPARENT_8_BIT)
			*maskBits = 0x34;					//	Pure green in BeOS palette.
		else
			*maskBits = B_TRANSPARENT_8_BIT;
	}
	greenMask->Unlock();
}

//	Load a bitmap from a given resource file.

PrefilledBitmap* LoadBitmapFromResource(BFile& file, int32 ID, uint32 type)
{
	BResources resFile(&file);	
	size_t bitmapSize;
	void *bitmapData = (void*)resFile.LoadResource(type, ID, &bitmapSize);
	if (!bitmapData)
		return 0;
	BMessage msg;
	msg.Unflatten((const char*)bitmapData);
	return new PrefilledBitmap(&msg);
}
