// Author: Dario Accornero <adario@cs.bu.edu>
// BeDev ID: 10183
// Project: BeScript
// Module:
// Error handling and file management.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

// System headers.

#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>

// Required headers.

#include "errormsg.h"

// GLOBAL DATA:
// Error encountered flag.

bool anyErrors = false;

// Reading position for the lexer.

int EM_tokPos = 0;

// EXTERNAL DECLARATIONS:
// Input file and restarting calls.

extern FILE*	yyin;
extern void		yyrestart(FILE* input_file);
extern int		yywrap(void);

// LOCAL CONSTANTS:
// Maximum error line string length.

const int kMaxErrLineLen = 1024;

// LOCAL STRUCTURES:
// Line tracking for lexical analysis.

typedef struct intList_ {
  int             i;
  struct intList_ *rest;
} *IntList;

// LOCAL DATA:
// Current file name and line/parse position.

static string  fileName = "";
static int     lineNum = 1;
static IntList linePos = NULL;

// LOCAL FUNCTIONS:
// Line list constructor.

static IntList intList(int i, IntList rest);
static IntList intList(int i, IntList rest) 
{
  IntList l = (IntList)malloc(sizeof(*l));
  l->i = i;
  l->rest = rest;
  return l;
}

// GLOBAL FUNCTIONS:
// Advance to next line position.

void EM_newline(void)
{
  ++lineNum;
  linePos = intList(EM_tokPos, linePos);
}

// Display an error message.

void EM_error(int pos, const char* message, ...)
{
  va_list ap;
  IntList lines = linePos; 
  int     num = lineNum;
  char    buf[kMaxErrLineLen];

  // Locate correct position in file.

  anyErrors = true;
  while (lines && lines->i >= pos) 
  {
    lines = lines->rest;
    --num;
  }

  // Possibly display file info.

  const char* str = fileName.c_str();
  if (str && strlen(str))
    fprintf(stderr, "%s: ", str);
  if (lines)
    fprintf(stderr, "line %d, char %d: ", num, pos-lines->i);

  // Display error message.

  va_start(ap, message);
  vfprintf(stderr, message, ap);
  va_end(ap);
  fputc('\n', stderr);

  // Possibly display the offending position in the file.

  if (yyin)
  {
    // Locate position and read line.

    int offset,
        curChar;

    fseek(yyin, lines->i, SEEK_SET);
    fgets(buf, kMaxErrLineLen, yyin);
    fprintf(stderr, "%s", buf);

    // Locate exactly current token position and display
    // a marker behind it.

    for (curChar = 0, offset = pos - lines->i; offset > 0; curChar++)
    {
      if (buf[curChar] == '\t')
      {
		fputc('\t', stderr);
		offset--;
      }
      else
      {
		fputc(' ', stderr);
		offset--;
      }
    }
    fprintf(stderr, "^^^\n");
  }
}

// Display an error message with source file and line information.
// This is identical to the above function apart from the two pieces
// of information mentioned above.

void EM_debugError(int line, const char* source, int pos, const char* message, ...)
{
  va_list ap;
  IntList lines = linePos; 
  int     num = lineNum;
  char    buf[kMaxErrLineLen];

  // Display source file information.

  fprintf(stderr, "(in %s line %d) ", source, line);

  // Locate correct position in file.

  anyErrors = true;
  while (lines && lines->i >= pos) 
  {
    lines = lines->rest;
    num--;
  }

  // Possibly display file info.

  const char* str = fileName.c_str();
  if (str && strlen(str))
    fprintf(stderr, "%s: ", str);
  if (lines)
    fprintf(stderr, "line %d, char %d: ", num, pos-lines->i);

  // Display error message.

  va_start(ap, message);
  vfprintf(stderr, message, ap);
  va_end(ap);
  fputc('\n', stderr);

  // Possibly display the offending position in the file.

  if (yyin)
  {
    // Locate position and read line.

    int offset,
        curChar;

    fseek(yyin, lines->i, SEEK_SET);
    fgets(buf, kMaxErrLineLen, yyin);
    fprintf(stderr, "%s", buf);

    // Locate exactly current token position and display
    // a marker behind it.

    for (curChar = 0, offset = pos - lines->i; offset > 0; curChar++)
    {
      if (buf[curChar] == '\t')
      {
		fputc('\t', stderr);
		offset--;
      }
      else
      {
		fputc(' ', stderr);
		offset--;
      }
    }
    fprintf(stderr, "^^^\n");
  }
}

// Support to open multiple files.
// Needed in the debugging phase.

void EM_open(const char* fname)
{
  // Reset error flag, assign file name, and reset lexer.

  anyErrors = false;
  fileName = fname;
  EM_reset();

  // Try to open the lexer input file.

  yyin = fopen(fname, "r");
  if (!yyin)
  {
    EM_error(0, "cannot open file");
    exit(1);
  }
}

// Reset lexer to beginning of file.

void EM_reset(void)
{
  // Discard the line tracking list.

  IntList lines = linePos;
  while (lines)
  {
    IntList toFree = lines;
    lines = lines->rest;
    free(toFree);
  }

  // Rebuild tracking list.

  lineNum = 1;
  linePos = intList(0, NULL);
  EM_tokPos = 0;

  // Reset the lex buffers, but only if a file has been
  // opened.

  if (yyin)
  {
    yyrestart(yyin);
    (void)yywrap();
  }
}

// Close current file and reset lexer.

void EM_close(void)
{
  // Reset the lex buffers and the input file.

  EM_reset();
  if (yyin)
  {
	  fclose(yyin);
	  yyin = NULL;
  }
}
